Imports System.Collections
Imports System.Threading


Public Class LayoutData
    'Class to store and manage All Layout Data and Assist with Object Retrieval
    '******************
    ' LAYOUT DATA
    '******************
    Public MaxRPCBytes As Integer = -1
    Private HighestUsedRPCBit As Integer = 0
    Private RPCRegister As SortedList = New SortedList(20)

    Public ShuntSig As ArrayList = New ArrayList(20)
    Public Sig2Asp As ArrayList = New ArrayList(20)
    Public Sig3Asp As ArrayList = New ArrayList(20)
    Public Sig4Asp As ArrayList = New ArrayList(20)
    Public IsolatedExits As ArrayList = New ArrayList(20)
    Public PlainTrack As ArrayList = New ArrayList(20)
    Public Points As ArrayList = New ArrayList(20)
    Public TkSec As ArrayList = New ArrayList(20)
    Public MiscIO As ArrayList = New ArrayList(20)
    Public MiscOP As ArrayList = New ArrayList(10)
    Public PKey As ArrayList = New ArrayList(20)

    Public Devices As ArrayList = New ArrayList(11) 'Holds Arraylists above containing each type.

    Public NameIndex As SortedList = New SortedList(100) 'List of names in use

    Public Routes As SortedList = New SortedList(100)
    Public RoutesIndex As SortedList = New SortedList(100)

    Public Requests As SortedList = New SortedList(100)

    Public DeviceIndex As Integer(,,) = New Integer(Max_X - 1, Max_Y - 1, 3) {}

    Private Signals As ISignal()
    Private Miscs As IMisc()

    '******************
    'FUNCTIONS
    '******************

    ' Initialises Data on Instance of Layout Data being created
    Public Sub New()
        Dim x, y As Integer
        ' Fills Devices Array List with Array Lists of Devices

        Devices.Insert(0, ShuntSig)
        Devices.Insert(1, Sig2Asp) 'Redundant
        Devices.Insert(2, Sig2Asp)
        Devices.Insert(3, Sig2Asp) 'Redundant
        Devices.Insert(4, Sig3Asp)
        Devices.Insert(5, Sig2Asp) 'Redundant
        Devices.Insert(6, Sig4Asp)
        Devices.Insert(7, Sig2Asp) 'Redundant
        Devices.Insert(8, IsolatedExits)
        Devices.Insert(9, PlainTrack)
        Devices.Insert(10, Points)
        Devices.Insert(11, TkSec)
        Devices.Insert(12, MiscIO)
        Devices.Insert(13, PKey)
        Devices.Insert(14, MiscOP)

        ' Place -1 (EMPTY) code into Object type for each cell
        ' Initialise Locations to 0
        For x = 1 To Max_X
            For y = 1 To Max_Y
                DeviceIndex(x - 1, y - 1, 0) = -1 'Device type code to get right storrage arraylist from Devices Arraylist
                DeviceIndex(x - 1, y - 1, 1) = 0 'Storage position within specific Arraylist
            Next
        Next

    End Sub

    '***************
    'DATA VALIDATION
    '***************
    'Validation of ALL Interlocking Data and Check of Point Keys
    'Must be Successful to Leave Design Mode
    Public Function ValidateAllData() As String
        Dim errorreport As String = ""

        'Validate Point Keys
        errorreport &= ValidateAllPointKeys()
        'Validate Routes
        errorreport &= ValidateAllRoutes()
        'Validate Routes
        errorreport &= ValidateAllRequests()
        'Validate Point IL Conditions
        errorreport &= ValidateAllPointIL()
        'Validate Signal IL Conditions
        errorreport &= ValidateAllSignalIL()
        'Validate Calculated Misc Outputs
        errorreport &= ValidateAllMiscOPIL()
        If MaxRPCBytes = -1 Then errorreport &= "RPC Hardware Configuration Not Entered"
        Return errorreport
    End Function

    Public Sub DisplayDataErrors()
        Dim errorreport As String = ValidateAllData()
        If errorreport <> "" Then
            Dim FaultsForm As New frmErrorReport()
            FaultsForm.ErrorBox.Text = errorreport
            FaultsForm.ErrorBox.SelectionLength = 0
            FaultsForm.ShowDialog()
            'Fault Form On Screen
            FaultsForm.Close()
            FaultsForm = Nothing
        End If
    End Sub

    Public Function ValidateAllRoutes() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To Data.Routes.Count
            errorreport &= CType(Data.Routes.GetByIndex(i - 1), Route).ValidateILData
        Next
        Return errorreport
    End Function

    Public Function ValidateAllRequests() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To Data.Requests.Count
            errorreport &= CType(Data.Requests.GetByIndex(i - 1), Request).ValidateILData
        Next
        Return errorreport
    End Function

    Public Function ValidateAllSignalIL() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To Data.ShuntSig.Count
            errorreport &= CType(Data.ShuntSig(i - 1), SignalShunt).ValidateILData()
        Next
        For i = 1 To Data.Sig2Asp.Count
            errorreport &= CType(Data.Sig2Asp(i - 1), Signal2Aspect).ValidateILData()
        Next
        For i = 1 To Data.Sig3Asp.Count
            errorreport &= CType(Data.Sig3Asp(i - 1), Signal3Aspect).ValidateILData()
        Next
        For i = 1 To Data.Sig4Asp.Count
            errorreport &= CType(Data.Sig4Asp(i - 1), Signal4Aspect).ValidateILData()
        Next
        Return errorreport
    End Function

    Public Function ValidateAllPointIL() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To Data.Points.Count
            errorreport &= CType(Data.Points(i - 1), Points).ValidateILData()
        Next
        Return errorreport
    End Function

    Public Function ValidateAllPointKeys() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To PKey.Count
            If CheckReference(PKey(i - 1).PointID) <> 10 Then
                errorreport &= "Point Key at (" & PKey(i - 1).X_Pos & ", " & PKey(i - 1).Y_Pos & ") references Invalid Point, " & PKey(i - 1).PointID & vbCrLf
            End If
        Next
        Return errorreport
    End Function

    Public Function ValidateAllMiscOPIL() As String
        Dim i As Integer
        Dim errorreport As String = ""
        For i = 1 To MiscOP.Count
            errorreport &= CType(Data.MiscOP(i - 1), MiscOutput).ValidateILData()
        Next
        Return errorreport
    End Function

    '************************
    'DATA RETRIEVAL FUNCTIONS
    '************************
    '
    Public Function GetDeviceType(ByVal x As Integer, ByVal y As Integer) As Integer
        Return DeviceIndex(x, y, 0)
    End Function

    Private Function GetDeviceObjects(ByVal x As Integer, ByVal y As Integer) As Object
        If GetDeviceType(x, y) <> -1 Then
            Return Devices(GetDeviceType(x, y))
        Else
            Return Nothing
            MessageBox.Show("GetDeviceObjects Function Failed - No Object exists at this cell", mrcc)
        End If
    End Function

    Public Function GetDevice(ByVal x As Integer, ByVal y As Integer) As Object
        Dim listofobjects As Object = GetDeviceObjects(x, y)
        Return listofobjects(DeviceIndex(x, y, 1))
    End Function

    Public Function GetObject(ByVal Reference As String) As Object
        Dim x As Integer = CType(NameIndex(Reference), Point).X()
        Dim y As Integer = CType(NameIndex(Reference), Point).Y()
        Return GetDevice(x, y)
    End Function

    Public Function GetDeviceLocation(ByVal x As Integer, ByVal y As Integer) As Integer
        Return DeviceIndex(x, y, 1)
    End Function

    Public Overloads Function GetSymbolCode(ByVal x As Integer, ByVal y As Integer) As Integer
        Return GetDevice(x, y).Symbol
    End Function

    Public Overloads Function GetSymbolCode(ByVal Name As String) As Integer
        Dim xy As Point = NameIndex(Name)
        Return GetDevice(xy.X, xy.Y).symbol
    End Function

    Public Function GetTKSection(ByVal Name As String) As TrackSection
        Return CType(GetByName(Name), TrackSection)
    End Function

    Public Function GetPoint(ByVal Name As String) As Points
        Return CType(GetByName(Name), Points)
    End Function

    Public Function GetStaticMisc(ByVal Name As String) As StaticMisc
        Return CType(GetByName(Name), StaticMisc)
    End Function

    Public Function GetMiscOP(ByVal Name As String) As MiscOutput
        Return CType(GetByName(Name), MiscOutput)
    End Function

    'Function to Shuffle up Object Locations in index after ArrayList.Remove
    Public Function RemoveFromIndex(ByVal type As Integer, ByVal location As Integer)
        Dim x As Integer = 0
        Dim y As Integer = 0
        For x = 0 To Max_X - 1
            For y = 0 To Max_Y - 1
                If Data.DeviceIndex(x, y, 0) = type Then
                    If Data.DeviceIndex(x, y, 1) > location Then Data.DeviceIndex(x, y, 1) -= 1
                End If
            Next
        Next
    End Function

    '******
    'ROUTES
    '******
    'Add Route
    Public Sub AddRoute(ByVal ID As String, ByVal RouteData As Route)
        Routes.Add(ID, RouteData)
        RoutesIndex.Add(MakeNX(RouteData.EntrySig, RouteData.ExitSig), RouteData.ID)
    End Sub

    'Remove Route
    Public Sub RemoveRoute(ByVal ID As String)
        Routes.Remove(ID)
        RoutesIndex.RemoveAt(RoutesIndex.IndexOfValue(ID))
    End Sub

    'Verify Route Exists
    Public Overloads Function CheckRouteExists(ByVal routeid As String) As Boolean
        Return Routes.ContainsKey(routeid)
    End Function

    Public Overloads Function CheckRouteExists(ByVal EntrySig As String, ByVal ExitSig As String) As Boolean
        Dim tempnx As String = ""
        tempnx = MakeNX(EntrySig, ExitSig)
        Return RoutesIndex.ContainsKey(tempnx)
    End Function

    'Get Route
    Public Function GetRoute(ByVal ID) As Route
        Return CType(Data.Routes(ID), Route)
    End Function

    'GetRoute by NX
    Public Function GetRouteByNX(ByVal NX As String) As Route
        Return CType(Routes(CType(RoutesIndex(NX), String)), Route)
    End Function

    'Set Auto Routes
    Public Sub SetAutoRoutes()
        Dim i As Integer
        For i = 1 To Data.Routes.Count
            CType(Data.Routes.GetByIndex(i - 1), Route).SetAutoRoutes()
        Next
    End Sub

    '********
    'REQUESTS
    '********
    'Add Route
    Public Sub AddRequest(ByVal ID As String, ByVal RequestData As Request)
        Requests.Add(ID, RequestData)
    End Sub

    'Remove Request
    Public Sub RemoveRequest(ByVal ID As String)
        Requests.Remove(ID)
    End Sub

    'Verify Request Exists
    Public Function CheckRequestExists(ByVal Requestid As String) As Boolean
        Return Requests.ContainsKey(Requestid)
    End Function

    'Get Request
    Public Function GetRequest(ByVal ID) As Request
        Return CType(Data.Requests(ID), Request)
    End Function

    '************
    'MISC DEVICES
    '************
    Public Sub CreateMiscInterface()
        Dim NumMiscs As Integer = MiscIO.Count + MiscOP.Count
        Dim NewMiscs As IMisc() = New IMisc(NumMiscs - 1) {}
        Dim i, j As Integer
        i = 0
        For j = 0 To MiscIO.Count - 1
            NewMiscs(i + j) = MiscIO(j)
        Next
        i += j
        For j = 0 To MiscOP.Count - 1
            NewMiscs(i + j) = MiscOP(j)
        Next

        Miscs = NewMiscs
    End Sub

    Public Sub UpdateMiscs()
        Dim misc As IMisc
        For Each misc In Miscs
            misc.Update()
        Next
    End Sub

    Public Sub MiscsToDesignMode()
        Dim misc As IMisc
        For Each misc In Miscs
            misc.EnterDesignMode()
        Next
    End Sub

    Public Sub MiscsToTestMode()
        Dim misc As IMisc
        For Each misc In Miscs
            misc.EnterTestMode()
        Next
    End Sub

    Public Sub MiscsToOperateMode()
        Dim misc As IMisc
        For Each misc In Miscs
            misc.EnterOperateMode()
        Next
    End Sub

    Public Function GetMiscState(ByVal MiscID As String) As Boolean
        Dim result As Boolean = False
        Dim misc As IMisc
        For Each misc In Miscs
            If misc.ID = MiscID Then Return misc.State
            'Jumps Out of Loop and stops searching if found
        Next
        'Catch All Returns a result
        Return result
    End Function

    '*******
    'SIGNALS
    '*******
    Public Sub CreateSignalInterface()
        Dim NumSignals As Integer = ShuntSig.Count + Sig2Asp.Count + Sig3Asp.Count + Sig4Asp.Count + IsolatedExits.Count
        Dim NewSignals As ISignal() = New ISignal(NumSignals - 1) {}
        Dim i, j As Integer
        i = 0
        For j = 0 To ShuntSig.Count - 1
            NewSignals(i + j) = ShuntSig(j)
        Next
        i += j
        For j = 0 To Sig2Asp.Count - 1
            NewSignals(i + j) = Sig2Asp(j)
        Next
        i += j
        For j = 0 To Sig3Asp.Count - 1
            NewSignals(i + j) = Sig3Asp(j)
        Next
        i += j
        For j = 0 To Sig4Asp.Count - 1
            NewSignals(i + j) = Sig4Asp(j)
        Next
        i += j
        For j = 0 To IsolatedExits.Count - 1
            NewSignals(i + j) = IsolatedExits(j)
        Next

        Signals = NewSignals
    End Sub

    Public Sub AspectSequencer()
        Dim Sig As ISignal
        For Each Sig In Signals
            Sig.UpdateAspect()
        Next
    End Sub

    Public Function IsRouteSetFromSig(ByVal Signal As String) As Boolean
        Dim i As Integer = 0
        Dim USet As Boolean = False
        For i = 0 To Me.Routes.Count - 1
            If CType(Routes.GetByIndex(i), Route).EntrySig = Signal And CType(Routes.GetByIndex(i), Route).USet = True Then
                USet = True
            End If
        Next
        Return USet
    End Function

    Public Sub SignalsToDesignMode()
        Dim sig As ISignal
        For Each sig In Signals
            sig.EnterDesignMode()
        Next
    End Sub

    Public Sub SignalsToOperateMode()
        Dim Sig As ISignal
        For Each Sig In Signals
            Sig.EnterOperateMode()
        Next
    End Sub

    '**********************
    'DEVICE NAME MANAGEMENT
    '**********************

    'Register Name/Reference
    Public Sub RegisterName(ByVal name As String, ByVal location As Point)
        NameIndex.Add(name, location)
    End Sub

    'Remove Name/Reference
    Public Sub RemoveName(ByVal name As String)
        NameIndex.Remove(name)
    End Sub

    'Check Reference
    Public Function CheckReference(ByVal name As String) As Integer
        Dim loc As Point
        If name = "**AUTO**"" then" Then
            Return 99 'To prevent registering a device named **AUTO**
        Else
            If NameIndex.Contains(name) Then
                loc = CType(NameIndex(name), Point)
                Return GetDeviceType(loc.X, loc.Y)
            Else
                Return -1
            End If
        End If
    End Function

    'Get Object By Name
    Public Function GetByName(ByVal name As String) As Object
        Dim loc As Point
        If NameIndex.Contains(name) Then
            loc = CType(NameIndex(name), Point)
            Return GetDevice(loc.X, loc.Y)
        End If
    End Function

    'Get Name by XY
    Public Function GetByXY(ByVal x As Integer, ByVal y As Integer) As String
        Dim loc As Point = MakePoint(x, y)
        Dim name As String = ""
        If NameIndex.ContainsValue(loc) Then
            name = NameIndex.GetKey(NameIndex.IndexOfValue(loc))
        End If
        Return name
    End Function

    '*****************************
    'RPC BIT ALLOCATION MANAGEMENT
    '*****************************
    '
    'Validate Bit
    Public Function ValidateRPCBit(ByVal bit As Integer) As Boolean
        If MaxRPCBytes = -1 Then
            MessageBox.Show("You must Enter the RPC Configuration before assigning Bits", mrcc)
            Return False
        Else
            If bit > ((MaxRPCBytes * 8) - 1) Then
                MessageBox.Show("Only Bits between 0 and " & MaxRPCBytes * 8 - 1 & " are available with current RPC Configuration." & vbCrLf & "Alter RPC Configuration if this is not correct.", mrcc)
                Return False
            Else
                Return True
            End If
        End If
    End Function

    'Need to Keep a Count of Allocations (Input or output doesn't matter)
    'to ensure that Highest Used is reduced correctly.

    'Register Bit
    Public Sub RegisterRPCBit(ByVal Bit As Integer)
        If RPCRegister.ContainsKey(Bit) Then
            'Increment Number of Alocations
            RPCRegister(Bit) += 1
        Else
            'Add Bit to Register
            RPCRegister.Add(Bit, 1)
            'If Highest Allocated, Set HighestUsed 
            If Bit > HighestUsedRPCBit Then HighestUsedRPCBit = Bit
        End If
    End Sub

    'DeRegister Bit
    Public Sub DeRegisterRPCBit(ByVal bit As Integer)
        If RPCRegister(bit) = 1 Then
            'Remove from Register
            RPCRegister.Remove(bit)
            'If this was highest, get next highest for HighestUsed
            If HighestUsedRPCBit = bit Then HighestUsedRPCBit = RPCRegister.GetKey(RPCRegister.Count - 1)
        Else
            'Decrement number of Allocations
            RPCRegister(bit) -= 1
        End If
    End Sub

    ' Returns Highest RPC Address entered - Used to stop Reducing RPC Bytes if Bit is being used
    Public ReadOnly Property GetHighestAddress() As Integer
        Get
            Return (HighestUsedRPCBit)
        End Get
    End Property

    '****************
    'DATA SAVING
    '****************
    Public Function SaveData() As String
        Dim n As Integer = 0

        Dim data As String = ""
        data += "MRCCC Layout File" & vbCrLf
        data += "=================" & vbCrLf & vbCrLf
        data += "Automatically Generated - Do NOT attempt to Alter Manually." & vbCrLf & vbCrLf
        data += "MRCCC Version : " & MRCCCVersion & vbCrLf & vbCrLf
        data += "Max RPC Bytes : " & MaxRPCBytes & vbCrLf & vbCrLf
        data += "!TRACK SECTIONS : " & TkSec.Count & vbCrLf
        data += " **************" & vbCrLf & vbCrLf
        If TkSec.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "Ref" & vbTab & "Has TC?" & vbTab & "TCRef" & vbTab & "Num Elements" & vbCrLf
            data += vbCrLf
            For n = 0 To TkSec.Count - 1
                data += CType(TkSec(n), TrackSection).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!PLAIN TRACK : " & PlainTrack.Count & vbCrLf
        data += " ***********" & vbCrLf & vbCrLf
        If PlainTrack.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "Sym" & vbTab & "TkAss?" & vbTab & "TkSec" & vbTab & "PntAss?" & vbTab & "Ref 1" & vbTab & "N/R" & vbTab & "Ref 2" & vbTab & "N/R" & vbTab & "Ref 3" & vbTab & "N/R" & vbCrLf
            data += vbCrLf
            For n = 0 To PlainTrack.Count - 1
                data += CType(PlainTrack(n), PlainTrack).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!POINTS : " & Points.Count & vbCrLf
        data += " ******" & vbCrLf & vbCrLf
        If Points.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "Ref" & vbTab & "RPC" & vbTab & "Sym" & vbTab & "TkAss?" & vbTab & "TkSec" & vbTab & "PntAss?" & vbTab & "Ref 1" & vbTab & "N/R" & vbTab & "Ref 2" & vbTab & "N/R" & vbTab & "Pair" & vbTab & "SwapNR" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Num Point Associations" & vbCrLf
            data += vbCrLf
            For n = 0 To Points.Count - 1
                data += CType(Points(n), Points).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!POINT KEYS : " & PKey.Count & vbCrLf
        data += " **********" & vbCrLf & vbCrLf
        If PKey.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "Ref" & vbCrLf
            data += vbCrLf
            For n = 0 To PKey.Count - 1
                data += CType(PKey(n), PointKey).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!EXITS : " & IsolatedExits.Count & vbCrLf
        data += " *****" & vbCrLf & vbCrLf
        If IsolatedExits.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Deflt" & vbTab & "Sym" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Auto?" & vbTab & "Num Routes from Sig" & vbCrLf
            data += vbCrLf
            For n = 0 To IsolatedExits.Count - 1
                data += CType(IsolatedExits(n), IsolatedExit).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!SHUNT SIGNALS : " & ShuntSig.Count & vbCrLf
        data += " *************" & vbCrLf & vbCrLf
        If ShuntSig.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Deflt" & vbTab & "Sym" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Auto?" & vbTab & "Mplex?" & vbTab & "RPC1" & vbTab & "RPC2" & vbTab & "NoRed?" & vbTab & "Num Routes from Sig" & vbCrLf
            data += vbCrLf
            For n = 0 To ShuntSig.Count - 1
                data += CType(ShuntSig(n), SignalShunt).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!2 ASPECT SIGNALS : " & Sig2Asp.Count & vbCrLf
        data += " ****************" & vbCrLf & vbCrLf
        If Sig2Asp.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Deflt" & vbTab & "Sym" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Auto?" & vbTab & "Mplex?" & vbTab & "RPC1" & vbTab & "RPC2" & vbTab & "Num Routes from Sig" & vbCrLf
            data += vbCrLf
            For n = 0 To Sig2Asp.Count - 1
                data += CType(Sig2Asp(n), Signal2Aspect).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!3 ASPECT SIGNALS : " & Sig3Asp.Count & vbCrLf
        data += " ****************" & vbCrLf & vbCrLf
        If Sig3Asp.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Deflt" & vbTab & "Sym" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Auto?" & vbTab & "Mplex?" & vbTab & "RPC1" & vbTab & "RPC2" & vbTab & "RPC3" & vbTab & "Num Routes from Sig" & vbCrLf
            data += vbCrLf
            For n = 0 To Sig3Asp.Count - 1
                data += CType(Sig3Asp(n), Signal3Aspect).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!4 ASPECT SIGNALS : " & Sig4Asp.Count & vbCrLf
        data += " ****************" & vbCrLf & vbCrLf
        If Sig4Asp.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Deflt" & vbTab & "Sym" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "Auto?" & vbTab & "Mplex?" & vbTab & "RPC1" & vbTab & "RPC2" & vbTab & "RPC3" & vbTab & "RPC4" & vbTab & "Num Routes from Sig" & vbCrLf
            data += vbCrLf
            For n = 0 To Sig4Asp.Count - 1
                data += CType(Sig4Asp(n), Signal4Aspect).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!MISC INPUT : " & MiscIO.Count & vbCrLf
        data += " **********" & vbCrLf & vbCrLf
        If MiscIO.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Type" & vbTab & "RPC Bit" & vbCrLf
            data += vbCrLf
            For n = 0 To MiscIO.Count - 1
                data += CType(MiscIO(n), StaticMisc).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!MISC OUTPUT : " & MiscOP.Count & vbCrLf
        data += " ***********" & vbCrLf & vbCrLf
        If MiscOP.Count > 0 Then
            data += "x" & vbTab & "y" & vbTab & "ID" & vbTab & "Type" & vbTab & "RPC Bit" & vbTab & "AND?" & vbTab & "1?" & vbTab & "1 Type" & vbTab & "1 ID" & vbTab & "1 Cond" & vbTab & "2?" & vbTab & "2 Type" & vbTab & "2 ID" & vbTab & "2 Cond" & vbTab & "3?" & vbTab & "3 Type" & vbTab & "3 ID" & vbTab & "3 Cond" & vbTab & "4?" & vbTab & "4 Type" & vbTab & "4 ID" & vbTab & "4 Cond" & vbTab & "5?" & vbTab & "5 Type" & vbTab & "5 ID" & vbTab & "5 Cond" & vbCrLf
            data += vbCrLf
            For n = 0 To MiscOP.Count - 1
                data += CType(MiscOP(n), MiscOutput).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!ROUTES : " & Routes.Count & vbCrLf
        data += " ******" & vbCrLf & vbCrLf
        If Routes.Count > 0 Then
            data += "ID" & vbTab & "Entry" & vbTab & "Exit" & vbTab & "MiscT" & vbTab & "MiscF" & vbTab & "ER Tk" & vbTab & "Clr/Occ" & vbTab & "A Button" & vbCrLf
            data += vbCrLf
            For n = 0 To Routes.Count - 1
                data += CType(Routes.GetByIndex(n), Route).SaveData() & vbCrLf
            Next
        End If
        data += vbCrLf
        data += "!REQUESTS : " & Requests.Count & vbCrLf
        data += " ********" & vbCrLf & vbCrLf
        If Requests.Count > 0 Then
            data += "ID" & vbTab & "Type1" & vbTab & "RPC Bit" & vbTab & "MiscID" & vbTab & "TF" & vbTab & "Type2" & vbTab & "ClkX" & vbTab & "ClkY" & vbTab & "ClkLR" & vbTab & "Route" & vbTab & "Set/Can" & vbTab & "Point" & vbTab & "NCR" & vbCrLf
            data += vbCrLf
            For n = 0 To Requests.Count - 1
                data += CType(Requests.GetByIndex(n), Request).SaveData() & vbCrLf
            Next
        End If
        Return data
    End Function
    Public Sub OpenData(ByVal data As String)
        Dim n As Integer = 0
        Dim i As Integer = 0
        Dim j As Integer = 0
        Dim ObjectType As Integer = 0
        Dim DataFile() As String = Split(data, vbCrLf)
        Dim NumLines As Integer = DataFile.GetLength(0)
        Dim Line As Integer = 0
        'Check Correct File Header
        If DataFile(0) = "MRCCC Layout File" Then
            'Check File Version
            Dim FileVer As String = Mid(DataFile(5), 17)
            Select Case (FileVer)
                Case FileVer = "999.9x"
                    'Sample of Version Specific
                Case Else
                    'Reconstruct Data
                    '****************
                    'RPC Bytes
                    MaxRPCBytes = ToInt(Mid(DataFile(7), 17))
                    'Track Sections
                    ObjectType = 11
                    Line = FindLine("TRACK SECTIONS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 19))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New TrackSection(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), CType(DataRecord(3), Boolean), ToInt(DataRecord(4))))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(4)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(4)))
                            If ToInt(DataRecord(5)) > 0 Then
                                For j = 1 To ToInt(DataRecord(5))
                                    Line += 1
                                    Dim xy() As String = Split(DataFile(Line), ",")
                                    GetTKSection(DataRecord(2)).AddTrackElement(ToInt(xy(0)), ToInt(xy(1)))
                                Next
                                Line = Line + 1
                            End If
                            Line = Line + 1
                        Next
                    End If
                    'Plain Track
                    ObjectType = 9
                    Line = FindLine("PLAIN TRACK", DataFile)
                    n = ToInt(Mid(DataFile(Line), 16))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New PlainTrack(ToInt(DataRecord(0)), ToInt(DataRecord(1)), ToInt(DataRecord(2)), CType(DataRecord(3), Boolean), DataRecord(4), CType(datarecord(5), Boolean), datarecord(6), datarecord(7), datarecord(8), datarecord(9), datarecord(10), datarecord(11)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            Line = Line + 1
                        Next
                    End If
                    'Points
                    ObjectType = 10
                    Line = FindLine("POINTS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 11))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New Points(ToInt(DataRecord(0)), ToInt(DataRecord(1)), ToInt(DataRecord(4)), CType(DataRecord(5), Boolean), DataRecord(6), datarecord(2), ToInt(datarecord(3)), CType(datarecord(7), Boolean), datarecord(8), datarecord(9), datarecord(10), datarecord(11), datarecord(12), CType(datarecord(13), Boolean), datarecord(14), datarecord(15)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(3)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(3)))
                            If ToInt(DataRecord(16)) > 0 Then
                                For j = 1 To ToInt(DataRecord(16))
                                    Line += 1
                                    Dim xy() As String = Split(DataFile(Line), ",")
                                    GetPoint(DataRecord(2)).AddTrackElement(ToInt(xy(0)), ToInt(xy(1)))
                                Next
                                Line = Line + 1
                            End If
                            Line = Line + 1
                        Next
                    End If
                    'Point Keys
                    ObjectType = 13
                    Line = FindLine("POINT KEYS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 15))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New PointKey(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            Line = Line + 1
                        Next
                    End If
                    'Exits
                    ObjectType = 8
                    Line = FindLine("EXITS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 10))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim Routelist As New ArrayList(10)
                            If ToInt(DataRecord(8)) > 0 Then
                                For j = 1 To ToInt(DataRecord(8))
                                    Line += 1
                                    Routelist.Add(DataFile(Line))
                                Next
                                Line = Line + 1
                            Else
                                Line = Line + 1
                            End If
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New IsolatedExit(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(DataRecord(3)), DataRecord(4), Routelist, datarecord(5), datarecord(6), CType(datarecord(7), Boolean)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            Line = Line + 1
                        Next
                    End If
                    'Shunt Signals
                    ObjectType = 0
                    Line = FindLine("SHUNT SIGNALS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 18))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim Routelist As New ArrayList(10)
                            If ToInt(DataRecord(12)) > 0 Then
                                For j = 1 To ToInt(DataRecord(12))
                                    Line += 1
                                    Routelist.Add(DataFile(Line))
                                Next
                                Line = Line + 1
                            Else
                                Line = Line + 1
                            End If
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New SignalShunt(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(DataRecord(3)), ToInt(DataRecord(4)), CType(datarecord(8), Boolean), ToInt(datarecord(9)), ToInt(datarecord(10)), CType(datarecord(11), Boolean), Routelist, datarecord(5), datarecord(6), CType(datarecord(7), Boolean)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(9)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(9)))
                            If ToInt(DataRecord(10)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(10)))
                            Line = Line + 1
                        Next
                    End If
                    '2 Aspect Signals
                    ObjectType = 2
                    Line = FindLine("2 ASPECT SIGNALS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 21))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim Routelist As New ArrayList(10)
                            If ToInt(DataRecord(11)) > 0 Then
                                For j = 1 To ToInt(DataRecord(11))
                                    Line += 1
                                    Routelist.Add(DataFile(Line))
                                Next
                                Line = Line + 1
                            Else
                                Line = Line + 1
                            End If
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New Signal2Aspect(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(DataRecord(3)), ToInt(DataRecord(4)), CType(datarecord(8), Boolean), ToInt(datarecord(9)), ToInt(datarecord(10)), Routelist, datarecord(5), datarecord(6), CType(datarecord(7), Boolean)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(9)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(9)))
                            If ToInt(DataRecord(10)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(10)))
                            Line = Line + 1
                        Next
                    End If
                    '3 Aspect Signals
                    ObjectType = 4
                    Line = FindLine("3 ASPECT SIGNALS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 21))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim Routelist As New ArrayList(10)
                            If ToInt(DataRecord(12)) > 0 Then
                                For j = 1 To ToInt(DataRecord(12))
                                    Line += 1
                                    Routelist.Add(DataFile(Line))
                                Next
                                Line = Line + 1
                            Else
                                Line = Line + 1
                            End If
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New Signal3Aspect(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(DataRecord(3)), ToInt(DataRecord(4)), CType(datarecord(8), Boolean), ToInt(datarecord(9)), ToInt(datarecord(10)), ToInt(datarecord(11)), Routelist, datarecord(5), datarecord(6), CType(datarecord(7), Boolean)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(9)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(9)))
                            If ToInt(DataRecord(10)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(10)))
                            If ToInt(DataRecord(11)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(11)))
                            Line = Line + 1
                        Next
                    End If
                    '4 Aspect Signals
                    ObjectType = 6
                    Line = FindLine("4 ASPECT SIGNALS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 21))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim Routelist As New ArrayList(10)
                            If ToInt(DataRecord(13)) > 0 Then
                                For j = 1 To ToInt(DataRecord(13))
                                    Line += 1
                                    Routelist.Add(DataFile(Line))
                                Next
                                Line = Line + 1
                            Else
                                Line = Line + 1
                            End If
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New Signal4Aspect(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(DataRecord(3)), ToInt(DataRecord(4)), CType(datarecord(8), Boolean), ToInt(datarecord(9)), ToInt(datarecord(10)), ToInt(datarecord(11)), ToInt(datarecord(12)), Routelist, datarecord(5), datarecord(6), CType(datarecord(7), Boolean)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(9)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(9)))
                            If ToInt(DataRecord(10)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(10)))
                            If ToInt(DataRecord(11)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(11)))
                            If ToInt(DataRecord(12)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(12)))
                            Line = Line + 1
                        Next
                    End If
                    'Misc Inputs
                    ObjectType = 12
                    Line = FindLine("MISC INPUT", DataFile)
                    n = ToInt(Mid(DataFile(Line), 15))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New StaticMisc(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(datarecord(3)), ToInt(datarecord(4))))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(4)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(4)))
                            Line = Line + 1
                        Next
                    End If
                    'Misc Outputs
                    ObjectType = 14
                    Line = FindLine("MISC OUTPUT", DataFile)
                    n = ToInt(Mid(DataFile(Line), 16))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim ObjectLocation As Integer = Devices(ObjectType).Add(New MiscOutput(ToInt(DataRecord(0)), ToInt(DataRecord(1)), DataRecord(2), ToInt(datarecord(3)), ToInt(datarecord(4)), CType(datarecord(5), Boolean), CType(datarecord(6), Boolean), ToInt(datarecord(7)), datarecord(8), datarecord(9), CType(datarecord(10), Boolean), ToInt(datarecord(11)), datarecord(12), datarecord(13), CType(datarecord(14), Boolean), ToInt(datarecord(15)), datarecord(16), datarecord(17), CType(datarecord(18), Boolean), ToInt(datarecord(19)), datarecord(20), datarecord(21), CType(datarecord(22), Boolean), ToInt(datarecord(23)), datarecord(24), datarecord(25)))
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 0) = ObjectType
                            DeviceIndex(ToInt(DataRecord(0)), ToInt(DataRecord(1)), 1) = ObjectLocation
                            NameIndex.Add(DataRecord(2), MakePoint(ToInt(DataRecord(0)), ToInt(DataRecord(1))))
                            If ToInt(DataRecord(4)) <> -1 Then RegisterRPCBit(ToInt(DataRecord(4)))
                            Line = Line + 1
                        Next
                    End If
                    'Routes
                    Line = FindLine("ROUTES", DataFile)
                    n = ToInt(Mid(DataFile(Line), 11))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            Dim NPointsList As New ArrayList(10)
                            Dim RPointsList As New ArrayList(10)
                            Dim ConflictList As New ArrayList(10)
                            Dim TkClrList As New ArrayList(10)

                            Line += 1
                            Dim TempData1() As String = Split(DataFile(Line), ",")
                            If ToInt(TempData1(1)) > 0 Then
                                For j = 2 To ToInt(TempData1(1) + 1)
                                    NPointsList.Add(TempData1(j))
                                Next
                            End If
                            Line += 1
                            Dim TempData2() = Split(DataFile(Line), ",")

                            If ToInt(TempData2(1)) > 0 Then
                                For j = 2 To ToInt(TempData2(1) + 1)
                                    RPointsList.Add(TempData2(j))
                                Next
                            End If
                            Line += 1
                            Dim TempData3() As String = Split(DataFile(Line), ",")
                            If ToInt(TempData3(1)) > 0 Then
                                For j = 2 To ToInt(TempData3(1) + 1)
                                    ConflictList.Add(TempData3(j))
                                Next
                            End If
                            Line += 1
                            Dim TempData4() As String = Split(DataFile(Line), ",")
                            If ToInt(TempData4(1)) > 0 Then
                                For j = 2 To ToInt(TempData4(1) + 1)
                                    TkClrList.Add(TempData4(j))
                                Next
                            End If

                            AddRoute(datarecord(0), New Route(datarecord(0), datarecord(1), datarecord(2), ConflictList, NPointsList, RPointsList, TkClrList, datarecord(3), datarecord(4), datarecord(5), datarecord(6), datarecord(7)))

                            Line += 2
                        Next
                    End If

                    'Requests
                    Line = FindLine("REQUESTS", DataFile)
                    n = ToInt(Mid(DataFile(Line), 13))
                    If n > 0 Then
                        Line = Line + 5 'Move to first data line
                        For i = 1 To n
                            Dim DataRecord() As String = Split(DataFile(Line), vbTab)
                            AddRequest(datarecord(0), New Request(datarecord(0), datarecord(1), datarecord(2), datarecord(3), datarecord(4), datarecord(5), datarecord(6), datarecord(7), datarecord(8), datarecord(9), datarecord(10), datarecord(11), datarecord(12)))
                            Line += 1
                        Next
                    End If
            End Select
        Else
            MessageBox.Show("Unable to Open" & vbCrLf & vbCrLf & "File " & fn & " is invalid or corrupted", mrcc)
        End If
    End Sub
    Private Function ToInt(ByVal str As String) As Integer
        Return CType(str, Integer)
    End Function
    Private Function FindLine(ByVal ID As String, ByVal DataFile() As String) As Integer
        Dim line As Integer = 0
        Dim result As Integer = -1
        Dim NumLines As Integer = DataFile.GetLength(0)
        Do Until line = NumLines Or result > -1
            If DataFile(line).StartsWith("!" & ID) Then result = line
            line += 1
        Loop
        Return result
    End Function
    '***************
    'DATA PROCESSING
    '***************
    '
    Public Sub Processing()
        Dim i As Integer

        'Admit One User Request
        If GUI.UserRequestQueue.Count > 0 Then
            Dim qst As UserRequest = CType(GUI.UserRequestQueue.Dequeue(), UserRequest)
            CType(GUI.GetUser(qst.pUser), User).MouseClick(qst)
        End If

        'Update Routes
        For i = 0 To Me.Routes.Count - 1
            CType(Routes.GetByIndex(i), Route).Update()
        Next

        'Update Track Sections
        For i = 0 To Me.TkSec.Count - 1
            CType(TkSec(i), TrackSection).Update()
        Next

        'Update Point Keys
        For i = 0 To Me.PKey.Count - 1
            CType(PKey(i), PointKey).Update()
        Next

        'Update Signals
        Data.AspectSequencer()

        'Update Misc
        Data.UpdateMiscs()

        'Process Requests
        For i = 0 To Me.Requests.Count - 1
            CType(Requests.GetByIndex(i), Request).Update()
        Next
    End Sub

End Class

'**********************
'DATA PROCESSING THREAD
'**********************
'
Class Processing
    Private StopThreadFlag As Boolean = False
    Private Process As Boolean = False
    Sub ProcessData()
        While StopThreadFlag = False
            If Process = True Then
                ProcessingFlag = True
                Data.Processing()
                ProcessingFlag = False
            End If
            If GUI.UserRequestQueue.Count = 0 Then Thread.Sleep(100)
        End While
    End Sub

    Public Property StopThread() As Boolean
        Get
            Return StopThreadFlag
        End Get
        Set(ByVal Value As Boolean)
            StopThreadFlag = Value
        End Set
    End Property

    Public Property Run() As Boolean
        Get
            Return Process
        End Get
        Set(ByVal Value As Boolean)
            Process = Value
        End Set
    End Property
End Class

